import sys
import json
import time

import requests
from requests.exceptions import ConnectTimeout

from .utils import requests_retry_session
from .exceptions import MissingArgumentError, TranscoderAlreadyExists
from .definitions import API_MAP, DEFAULT_CREATE_DATA
from .cli.definitions import REQUIRED_CREATE_ARGS


class PexipConnection:
    """Abstract class for a connection to Pexip Meeting Manager."""

    routes = API_MAP

    def __init__(self, args):
        """ Minimal Requirements to establish a remote connection. """
        self.url = args.manager_url
        self.username = args.auth_user
        self.password = args.auth_pass

        # Supress warnings if explicitly running with --insecure
        if args.insecure:
            requests.packages.urllib3.disable_warnings()
        # Flip --insecure (store_true) to Verify -> false
        self.verify = not args.insecure

        # Create Client with Auth
        self.client = requests.Session()
        self.client.auth = (self.username, self.password)

        # Store args for access later
        self.args = args

    def _filter_by(self, objs, key, val=None):
        """ Iterate over a list of dicts, search check each obj for key == val """
        for obj in objs:
            if obj.get(key) == val:
                return obj
        return None

    def _log_error(self, status, content):
        if status == 401:
            print(f"Error {status}: Are auth_user and auth_pass set correctly?")
        else:
            print(f"Error {status}: Content: {content}")

    def _get_config(self):
        raise NotImplementedError()


class PexipClient(PexipConnection):
    """ A Client for basic operations with a PexipManager """

    def _get_config(self, endpoint):
        """ Requests config attributes from the Pexip Meeting Manager. """
        route = self.routes[endpoint]
        response = self.client.get(f"{self.url}{route}", verify=self.verify)
        if not response.ok:
            self._log_error(response.status_code, response.content)

        data = json.loads(response.content)

        return data

    def _get_tls_config(self, tls_certificate_subject_name):
        """Fetch the ID of the supplied TLS Certificate Subject Name and return a
        string pointing at its full path on system."""

        data = self._get_config("tls_cert")
        tls_cert_id = self._filter_by(
            data.get("objects", {}), key="subject_name", val=tls_certificate_subject_name
        )
        if not tls_cert_id:
            raise Exception(
                f"Error: TLS Certificate ID with name '{tls_certificate_subject_name}' was unable to be found."
            )
        # e.g. '/api/admin/configuration/v1/tls_certificate/1/'
        return f"{self.routes['tls_cert']}{tls_cert_id['id']}/"

    def _get_system_location_id(self, system_location_name):
        """Fetch the ID of the supplied System Location and return a
        string pointing at its full path on system."""

        data = self._get_config("system_location")
        system_location_id = self._filter_by(
            data.get("objects", {}), key="name", val=system_location_name
        )
        if not system_location_id:
            raise Exception(
                f"Error: System Location with name '{system_location_name}' was unable to be found."
            )
        # e.g. '/api/admin/configuration/v1/system_location/1/'
        return f"{self.routes['system_location']}{system_location_id['id']}/"

    def _node_exists(self, hostname) -> bool:
        """ Check if a specific node exists within Manager's 'worker_vm' API """
        data = self._get_config("node")
        return bool(self._filter_by(data.get("objects", {}), key="name", val=hostname))

    def _sanity_check(self):
        for arg in REQUIRED_CREATE_ARGS:
            if not getattr(self.args, arg):
                raise MissingArgumentError(
                    f"Argument --{arg.replace('_', '-')} is required to create a node."
                )
        if self._node_exists(self.args.hostname):
            raise TranscoderAlreadyExists(
                f"Transcoder Node {self.args.hostname} already exists. Nothing to do."
            )


class PexipNode(PexipClient):
    """ A Pexip Transcoder Node. """

    def create(self):
        # Validate input data before POSTing config to manager
        self._sanity_check()

        tls_cert_path = self._get_tls_config(self.args.tls_certificate_subject_name)
        system_location_path = self._get_system_location_id(self.args.system_location_name)

        data = {
            "name": self.args.hostname,
            "hostname": self.args.hostname,
            "domain": self.args.domain,
            "address": self.args.private_ip,
            "static_nat_address": self.args.public_ip,
            "netmask": self.args.netmask,
            "gateway": self.args.gateway,
            "password": self.args.node_password,
            "tls_certificate": tls_cert_path,
            "system_location": system_location_path,
        }

        response = self.client.post(
            f"{self.url}/{self.routes['node']}",
            stream=True,
            data=json.dumps({**data, **DEFAULT_CREATE_DATA}),
        )

        if not response.ok:
            # TODO: if debug print(data) ?
            self._log_error(response.status_code, response.content)
            sys.exit(1)

        # TODO: Logging
        print(f"Successfully created configuration for {self.args.hostname}")
        self._provision(response.content)

    def _provision(self, content):
        start_time = time.time()
        try:
            print(f"Attempting to provision the node {self.args.hostname}")
            # Send request, wait up to 5s (timeout=5) for a response.
            # NOTE: This will not wait the full 5s if the host response with an error after, 2 seconds
            # NOTE: This is a freshly stood up node, it will have a self-signed certificate. Verify MUST be false.
            response = requests_retry_session().post(
                f"https://{self.args.private_ip}:8443/configuration/bootstrap",
                verify=False,
                headers={"Content-Type": "text/xml"},
                data=content,
                timeout=5,
            )
        except ConnectTimeout as e:
            print(
                f"Connection timed out when attempting to provision node: {self.args.hostname}. Error: {e}"
            )
            self._error_and_write_xml(content)
        except Exception as e:
            print(
                f"Unknown error when attempting to provision node: {self.args.hostname}. Error: {e}"
            )
            self._error_and_write_xml(content)
        else:
            if response.ok:
                print(f"Success: {response.status_code} status code.")
            else:
                print(
                    f"Error in response returned when provisioning node. Status code: {response.status_code}. Response Content: {response.content}."
                )
                self._error_and_write_xml(content)
        finally:
            end_time = time.time()
            print(f"Total time {end_time - start_time}")

    def _error_and_write_xml(self, content):
        with open(f"{self.args.hostname}.xml", "wb") as xml_file:
            print(f"Writing {self.args.hostname}.xml config to disk for retry.")
            xml_file.write(content)


class PexipManager(PexipConnection):
    """Class for controlling the Pexip Manager configuration.  """

    pass


class DataObject(dict):
    """ Simple object for dot walking """

    def __init__(self, *args, **kwargs):
        self.update(*args, **kwargs)

    def __getattr__(self, attr):
        return self.get(attr)

    def __setattr__(self, key, value):
        self.__setitem__(key, value)

    def __setitem__(self, key, value):
        self.__dict__.update({key: value})

    def __delattr__(self, item):
        self.__delitem__(item)

    def __delitem__(self, key):
        del self.__dict__[key]



import sys
import json
import argparse

from textwrap import dedent
from argparse import RawDescriptionHelpFormatter

from ..models import DataObject
from ..config import Config


class PexipHelpFormatter(RawDescriptionHelpFormatter):
    """A nicer help formatter.
    Help for arguments can be indented and contain new lines.
    It will be de-dented and arguments in the help
    will be separated by a blank line for better readability.
    """

    def __init__(self, max_help_position=6, *args, **kwargs):
        # A smaller indent for args help.
        kwargs["max_help_position"] = max_help_position
        super().__init__(*args, **kwargs)

    def _split_lines(self, text, width):
        text = dedent(text).strip() + "\n\n"
        return text.splitlines()


class PexipArgumentParser(argparse.ArgumentParser):
    """Adds additional logic to `argparse.ArgumentParser`.
    Handles all input (CLI args, file args, stdin), applies defaults,
    and performs extra validation.
    """

    def __init__(self, *args, formatter_class=PexipHelpFormatter, **kwargs):
        kwargs["add_help"] = True
        super().__init__(*args, formatter_class=formatter_class, **kwargs)
        self.cfg = None  # Represents the local config
        self.args = None  # Contains initial CLI data
        self.data = None  # Contains shaped CLI data
        self.action = None  # Represents the action to take
        self.hostname = None  # A 'hostname' or list of 'hostnames'

    def parse_args(self, cfg: Config, args=None, namespace=None) -> argparse.Namespace:
        self.cfg = cfg

        self.args = super().parse_args(args, namespace)
        try:
            self.action = self.args.subparser
        except AttributeError:
            super().print_help()
            sys.exit()

        self._apply_config_values()
        self._split_out_hostname()
        self._remerge_data()

        return (self.data, self.action)

    def _apply_config_values(self):
        """Iterate through values found in the config. If no CLI values were passed,
        set them to the configs value."""

        for key, value in self.cfg.items():
            cfg_key = key.lower()
            if not getattr(self.args, cfg_key):
                setattr(self.args, cfg_key, value)

    def _split_out_hostname(self):
        """ Attempt to parse hostname positional argument as JSON blob """
        try:
            # Attempt load the 'hostname' argument up as a JSON dictionary
            self.hostname = json.loads(self.args.hostname)
        except json.decoder.JSONDecodeError:
            self.hostname = self.args.hostname
        else:
            # Upon successful json.loads, remove the hostname attribute from self.args
            del self.args.hostname

    def _remerge_data(self):
        """ Merge in global argument data with every hostname object """
        if isinstance(self.hostname, list):
            # Create a list of DataObjects merged in with top level config values
            # NOTE: the 2nd merged value will win in the event of a key conflict.
            self.data = [DataObject({**vars(self.args), **hostname}) for hostname in self.hostname]
        else:
            self.data = self.args



import os
import sys
import json
import time
import requests

# Local import
from .utils import requests_retry_session

# Disable annoying warnings if we're knowingly setting SSL_VERIFY to False.
if not SSL_VERIFY:
    requests.packages.urllib3.disable_warnings()


class PexipClient:
    def __init__(self, args):
        # Pexip TLS Certificate / System Location
        self.tls_certificate_subject_name = args.tls_certificate_subject_name
        self.system_location_name = args.system_location_name

    def _get_tls_cert_id(self):
        data = self._get_pexip_config("/api/admin/configuration/v1/tls_certificate/")
        tls_cert_id = self._filter_by(
            data.get("objects", {}), key="subject_name", val=self.tls_certificate_subject_name
        )
        if not tls_cert_id:
            raise (
                "Error: TLS Certificate ID with name '{self.tls_certificate_subject_name}' was unable to be found."
            )
        return tls_cert_id["id"]

    def _get_system_location_id(self):
        data = self._get_pexip_config("/api/admin/configuration/v1/system_location/")
        system_location_id = self._filter_by(
            data.get("objects", {}), key="name", val=self.system_location_name
        )
        if not system_location_id:
            raise (
                "Error: System Location with name '{self.system_location_name}' was unable to be found."
            )
        return system_location_id["id"]

    def take_action(self, action):
        if action == "create":
            if not self.node_exists():
                xml_doc = self.create_node()
                self.provision_node(xml_doc)
            else:
                print(f"Configuration for '{self.hostname}' found at '{self.manager_url}'")
                print("No changes were made. Exiting.")
                sys.exit(0)
        elif action == "remove":
            self.remove_node()
        else:
            print(f"Action: {action} not yet implemented. Exiting.")

    def node_exists(self):
        data = self._get_pexip_config()
        node_existence = self._filter_by(data.get("objects", {}), key="name", val=self.hostname)
        if node_existence:
            return True
        else:
            return False



# These values are "required" to create a Pexip Transcoder within Manager
REQUIRED_CREATE_ARGS = [
    "hostname", 
    "domain", 
    "private_ip", 
    "public_ip", 
    "netmask", 
    "gateway", 
    "node_password", 
    "tls_certificate_subject_name", 
    "system_location_name"
]



import os
import json
import errno

from pathlib import Path
from typing import Union

DEFAULT_RELATIVE_CONFIG_HOME = Path(".pexip")

ENV_PEXIP_CONFIG_DIR = "PEXIP_CONFIG_DIR"


def get_default_config_dir() -> Path:
    """
    Find the default Pexip config directory.
    """

    # 1. Grab through Env
    env_config_dir = os.environ.get(ENV_PEXIP_CONFIG_DIR)
    if env_config_dir:
        return Path(env_config_dir)

    # 2 Check local directory to see if .pexip/ exists
    local_config_dir = Path.cwd() / DEFAULT_RELATIVE_CONFIG_HOME
    if local_config_dir.exists():
        return local_config_dir

    # 3 Check home directory to see if .pexip/ exists
    home_config_dir = Path.home() / DEFAULT_RELATIVE_CONFIG_HOME
    if home_config_dir.exists():
        return home_config_dir


DEFAULT_CONFIG_DIR = get_default_config_dir()


class ConfigFileError(Exception):
    pass


class BaseConfigDict(dict):
    name = None
    helpurl = None
    about = None

    def __init__(self, path: Path):
        super().__init__()
        self.path = path


class Config(BaseConfigDict):
    FILENAME = "config.json"

    def __init__(self, directory=DEFAULT_CONFIG_DIR):
        # if a directory isn't found, just return
        if not directory:
            return 

        self.directory = Path(directory)
        super().__init__(path=self.directory / self.FILENAME)
        try:
            with self.path.open("rt") as f:
                try:
                    data = json.load(f)
                except ValueError as e:
                    data = {}
        except IOError as e:
            data = {}

        self.update(data)



import os
import sys

from argparse import FileType, OPTIONAL, SUPPRESS, ZERO_OR_MORE

from .argparser import PexipArgumentParser

from textwrap import dedent

parser = PexipArgumentParser(
    prog="pexip",
    description="Pexip API Interface",
)

# Create Sub parsers
subparsers = parser.add_subparsers()

global_options = parser.add_argument_group(title="Global Options", description=None)

###
# Global Opts
###

global_options.add_argument(
    "--manager-url",
    help="""
    The URL of the Pexip Manager to interface with. This value will override 
    any MANAGER_URL values defined within the Pexip config.json file.
    """,
)
global_options.add_argument(
    "--insecure",
    action="store_true",
    help="""
    Allow insecure (non-TLS) connections with the Pexip Manager.
    """,
)

###
# Create
###
provision_create = subparsers.add_parser("create")
provision_create.set_defaults(subparser="create")

provision_create.add_argument(
    "--domain",
    help="""
    The domain of the new node.""",
)
provision_create.add_argument(
    "--private-ip",
    help="The private IP address of the Pexip node to provision.",
)
provision_create.add_argument(
    "--public-ip",
    help="The public IP address of the Pexip node to provision.",
)
provision_create.add_argument(
    "--netmask",
    default="255.255.255.0",
    help="The IP netmask of the new node. Default: '255.255.255.0'",
)
provision_create.add_argument(
    "--gateway",
    help="The Gateway IP for new node.",
)
provision_create.add_argument(
    "--node-password",
    default=os.environ.get("PEXIP_TRANSCODER_PASSWORD", "SuperSecr3t!"),
    help="The password to use for the new Transcoder node.",
)
provision_create.add_argument(
    "--tls-certificate-subject-name",
    help="""
    Subject name of the TLS Certificate to apply to the node.""",
)

###
# Delete
###
provision_delete = subparsers.add_parser("delete")
provision_delete.set_defaults(subparser="delete")

###
# Common Create / Delete Arguments
###

for subparser in [provision_create, provision_delete]:
    subparser.add_argument(
        "hostname",
        help="The hostname of the Pexip node to target.",
    )
    subparser.add_argument(
        "-e",
        "--environment",
        default="staging",
        help="""
        The deployment environment of the conference node. 
        Default: staging""",
    )
    subparser.add_argument(
        "-u",
        "--auth_user",
        default=os.environ.get("MEETING_MANAGER_USER", "admin"),
        help="""
        User to authenticate against meet manager with. 
        Default: 'admin'""",
    )
    subparser.add_argument(
        "-p",
        "--auth_pass",
        default=os.environ.get("MEETING_MANAGER_PASS", ""),
        help="""
        Password to auhenticate against meet manager with. 
        Default: env['MEETING_MANAGER_PASS']""",
    )
    subparser.add_argument(
        "--system-location-name",
        help="Name of the system location to create or delete conferencing node from.",
    )



import os
import sys
import json
import configparser

from .config import Config
from .models import DataObject, PexipNode
from .exceptions import TranscoderAlreadyExists


def create_node(transcoder):
    node = PexipNode(transcoder)
    try:
        node.create()
    except TranscoderAlreadyExists as msg:
        print(msg)


def create(args):
    """ Route Pexip Transcoder creation. """
    if isinstance(args, list):
        for arg in args:
            create_node(arg)
    else:
        create_node(args)


def delete(args):
    print("TODO: Implement Delete!")
    sys.exit()


def main(args=sys.argv, cfg=Config()):
    program_name, *args = args

    from .cli.args import parser

    data, action = parser.parse_args(args=args, cfg=cfg)

    actions = {
        "create": create,
        "delete": delete
    }

    actions[action](data)



import requests
from requests.adapters import HTTPAdapter
from requests.packages.urllib3.util.retry import Retry

# These globals are for 'requests_retry_session' in conjunction with sending requests to a
# freshly provisioning EC2 AMI (the conference node)
PROVISION_RETRIES = 10        # Make a total of int() requests.
PROVISION_BACKOFF_FACTOR = 1  # A backoff factor of 1 will look like 1, 2, 4, 8, 16 for 5 requests.

def requests_retry_session(
    retries=PROVISION_RETRIES,
    backoff_factor=PROVISION_BACKOFF_FACTOR,
    status_forcelist=(500, 502, 504),
    session=None,
):
    """ Create a requests 'session' object for multiple retries. """
    session = session or requests.Session()
    retry = Retry(
        total=retries,
        read=retries,
        connect=retries,
        backoff_factor=backoff_factor,
        status_forcelist=status_forcelist,
    )
    adapter = HTTPAdapter(max_retries=retry)
    session.mount("http://", adapter)
    session.mount("https://", adapter)
    return session



SYSTEM_SERVICES = ["dns", "ntp", "syslog"]
PLATFORM_SERVICES = ["license", "license_request", "tls"]
ALL_SERVICES = [SYSTEM_SERVICES, PLATFORM_SERVICES]

SYSTEM_MAP = {
    "dns": "/api/admin/configuration/v1/dns_server/",
    "ntp": "/api/admin/configuration/v1/ntp_server/",
    "syslog": "/api/admin/configuration/v1/syslog_server/",
    "proxy": "/api/admin/configuration/v1/http_proxy/",
    "smtp": "/api/admin/configuration/v1/snmp_network_management_system/",
    "vm_manager": "/api/admin/configuration/v1/host_system/",
    "static_route": "/api/admin/configuration/v1/static_route/",
}

PLATFORM_MAP = {
    "system_location": "/api/admin/configuration/v1/system_location/",
    "management": "/api/admin/configuration/v1/management_vm/",
    "node": "/api/admin/configuration/v1/worker_vm/",
    "license": "/api/admin/configuration/v1/licence/",
    "license_request": "/api/admin/configuration/v1/licence_request/",
    "ca_cert": "/api/admin/configuration/v1/ca_certificate/",
    "tls_cert": "/api/admin/configuration/v1/tls_certificate/",
    "csr": "/api/admin/configuration/v1/certificate_signing_request/",
    "global": "/api/admin/configuration/v1/global/",
    "diag_graphs": "/api/admin/configuration/v1/diagnostic_graphs/",
}

API_MAP = {**SYSTEM_MAP, **PLATFORM_MAP}

# Default data object for Node creation.
DEFAULT_CREATE_DATA = {
    "description": "Transcoding Node",
    "node_type": "CONFERENCING",
    "deployment_type": "MANUAL-PROVISION-ONLY",
}
